package jobqueue

import (
	"errors"
	"gorm.io/gorm"
)

func (s *JobSyncer) DeleteJob(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()

		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}

		if err := tx.Where("job_id = ?", permJob.GetID()).Delete(&JobStats{}).Error; err != nil {
			return err
		}

		if err := tx.Delete(&permJob).Error; err != nil {
			return err
		}

		return nil
	})
}

func (s *JobSyncer) ResetLogs(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		permJob := job.GetPersistence()
		if err := tx.Unscoped().Where("job_id = ?", permJob.GetID()).Delete(&JobLog{}).Error; err != nil {
			return err
		}

		return nil
	})
}

func (s *JobSyncer) ResetStats(job GenericJob) error {
	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	job.ResetStats()
	stats := job.GetStats()
	return s.manager.database.Transaction(func(tx *gorm.DB) error {
		return tx.Model(&JobStats{}).Where("job_id = ?", job.GetID()).Select("*").Omit("deleted_at", "created_at", "job_id").Updates(stats).Error

	})
}

func (s *JobSyncer) CreateOrUpdateJob(job GenericJob) error {

	s.mu.Lock()
	defer s.mu.Unlock()

	if s.manager == nil || s.manager.database == nil {
		return ErrNoDatabaseConnection
	}

	return createOrUpdateJob(job, s.manager.database)

}

func createOrUpdateJob(job GenericJob, db *gorm.DB) error {

	return db.Transaction(func(tx *gorm.DB) error {

		permJob := job.GetPersistence()

		memLogs := permJob.Logs
		permJob.Logs = nil

		var existingJob JobPersistence
		result := tx.Unscoped().Where("id = ?", permJob.GetID()).First(&existingJob)

		if result.Error != nil {
			if errors.Is(result.Error, gorm.ErrRecordNotFound) {
				err := tx.Create(&permJob).Error
				if err != nil {
					Trace("Error while creating job", "error", err)
					return err
				}
			} else {
				Trace("Error while creating job", "error", result.Error)
				return result.Error
			}
		} else {

			// remove deleted from
			r := tx.Unscoped().Model(&existingJob).Where("id = ?", existingJob.ID).Select("*").
				Omit("deleted_at", "created_at", "job_id").Update("deleted_at", nil)
			if r.Error != nil {
				Trace("Error while deleting job", "error", r.Error)
				return r.Error
			}

			// update scheduler
			resultStats := tx.Model(&existingJob).Where("id = ?", existingJob.ID).Select("*").
				Omit("deleted_at", "created_at", "job_id").UpdateColumns(SchedulerPersistence{
				Type:     "",
				Interval: 0,
				Spec:     "",
				Delay:    0,
				Event:    "",
				Time:     nil,
				Executed: false,
			})

			if resultStats.Error != nil {
				Trace("Error while updating job stats", "error", resultStats.Error)
				return resultStats.Error
			}

			r2 := tx.Model(&existingJob).Where("id = ?", existingJob.ID).Select("*").
				Omit("deleted_at", "created_at", "job_id").Updates(permJob)
			if r2.Error != nil {
				return r2.Error
			}
		}

		if tx.Error != nil {
			Trace("Error while updating job", "error", tx.Error)
			return tx.Error
		}

		tx.Model(&permJob.Stats).
			Select("*").Omit("deleted_at", "created_at", "job_id").
			UpdateColumns(permJob.Stats)

		if tx.Error != nil {
			Error("Error while updating job stats", "error", tx.Error)
			return result.Error
		}

		for i, _ := range memLogs {
			memLogs[i].LogID = 0
			_ = tx.Create(&memLogs[i])
			// no error handling, if it fails, it fails
		}

		return nil

	})

}
