package jobqueue

import (
	"errors"
	"strconv"
	"sync"
	"testing"
	"time"
)

func TestResolve(t *testing.T) {
	p := NewPromise[int](func(resolve func(int), reject func(error)) {
		time.Sleep(time.Second)
		resolve(42)
	})

	wg := sync.WaitGroup{}
	wg.Add(1)

	Then(p, func(value int) (string, error) {
		wg.Done()
		return "ok", nil
	}, func(e error) error {
		t.Fail()
		return nil
	})

	wg.Wait()
}

func TestReject(t *testing.T) {
	p := NewPromise[int](func(resolve func(int), reject func(error)) {
		reject(errors.New("rejected"))
	})

	wg := sync.WaitGroup{}
	wg.Add(1)

	Then(p, func(value int) (string, error) {
		t.Fail()
		return "", errors.New("should not be called")
	}, func(e error) error {
		if e.Error() != "rejected" {
			t.Fail()
		}
		wg.Done()
		return nil
	})
	wg.Wait()
}

func TestCatch(t *testing.T) {
	p := NewPromise[int](func(resolve func(int), reject func(error)) {
		reject(errors.New("catch error"))
	})

	wg := sync.WaitGroup{}
	wg.Add(1)

	Catch(p, func(e error) error {
		if e.Error() != "catch error" {
			t.Fail()
		}
		wg.Done()
		return nil
	})
	wg.Wait()
}

func TestChaining(t *testing.T) {
	p := NewPromise[int](func(resolve func(int), reject func(error)) {
		resolve(42)
	})

	wg := sync.WaitGroup{}
	wg.Add(1)

	p2 := Then(p, func(value int) (string, error) {
		return "ok", nil
	}, func(e error) error {
		t.Fail()
		return nil
	})
	Catch(p2, func(e error) error {
		t.Fail()
		return nil
	})

	Then(p2, func(value string) (int, error) {
		wg.Done()
		return 1, nil
	}, func(e error) error {
		t.Fail()
		return nil
	})

	wg.Wait()
}

func TestChainingWithMultipleThen(t *testing.T) {
	p := NewPromise[int](func(resolve func(int), reject func(error)) {
		resolve(42)
	})

	wg := sync.WaitGroup{}
	wg.Add(1)

	p2 := Then(p, func(value int) (string, error) {
		return "ok", nil
	}, func(e error) error {
		t.Fail()
		return nil
	})
	p3 := Then(p2, func(value string) (int, error) {
		wg.Done()
		return 1, nil
	}, func(e error) error {
		t.Fail()
		return nil
	})
	Catch(p3, func(e error) error {
		t.Fail()
		return nil
	})
	wg.Wait()
}

func TestThenWithManyInputs(t *testing.T) {
	// Testfälle mit verschiedenen Eingabetypen und -werten
	testCases := []struct {
		name     string
		input    interface{}
		expected interface{}
	}{
		// Erfolgreiche Versprechen mit verschiedenen Datentypen
		{"test int", 10, 10},
		{"test string", "Hallo Welt", "Hallo Welt"},
		{"test bool", true, true},
		{"test struct", struct{ Name string }{Name: "Bard"}, struct{ Name string }{Name: "Bard"}},

		// Erfolgreiche Versprechen mit komplexen Werten
		{"test int array", []int{}, []int{}},
		{"test string map", map[string]string{}, map[string]string{}},
		{"test struct pointer", &struct{ Name string }{Name: "Bard"}, &struct{ Name string }{Name: "Bard"}},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			p := NewPromise[any](func(resolve func(interface{}), reject func(error)) {
				resolve(tc.input)
			})

			wg := &sync.WaitGroup{}
			wg.Add(1)

			Then(p, func(value any) (interface{}, error) {

				switch value.(type) {
				case int:
					exp := tc.expected.(int)
					if value.(int) != exp {
						t.Fail()
					}
				case string:
					exp := tc.expected.(string)
					if value != exp {
						t.Fail()
					}

				case bool:
					exp := tc.expected.(bool)
					if value != exp {
						t.Fail()
					}

				case struct{ Name string }:
					exp := tc.expected.(struct{ Name string })
					if value != exp {
						t.Fail()
					}

				case []int:
					exp := tc.expected.([]int)
					if len(value.([]int)) != len(exp) {
						t.Fail()
					}

				case map[string]string:
					exp := tc.expected.(map[string]string)
					if len(value.(map[string]string)) != len(exp) {
						t.Fail()
					}

				case *struct{ Name string }:
					exp := tc.expected.(*struct{ Name string })
					convertValue := value.(*struct{ Name string })
					if convertValue.Name != exp.Name {
						t.Fail()
					}

				default:
					t.Fail()
				}

				wg.Done()
				return nil, nil
			}, func(e error) error {
				t.Fail()
				return nil
			})

			wg.Wait()

		})
	}
}

func TestPromiseSuccess(t *testing.T) {
	promise := NewPromise[int](func(resolve func(int), reject func(error)) {
		time.Sleep(100 * time.Millisecond)
		resolve(42)
	})

	resultPromise := Then(promise, func(value int) (string, error) {
		return "Value is: " + strconv.Itoa(value), nil
	}, func(err error) error {
		return err
	})

	<-resultPromise.done

	if resultPromise.err != nil {
		t.Errorf("Promise was rejected with error: %v", resultPromise.err)
	}

	if resultPromise.value != "Value is: 42" {
		t.Errorf("Promise was resolved with incorrect value: %v", resultPromise.value)
	}
}

func TestPromiseFailure(t *testing.T) {
	expectedError := errors.New("something went wrong")

	promise := NewPromise[int](func(resolve func(int), reject func(error)) {
		time.Sleep(100 * time.Millisecond)
		reject(expectedError)
	})

	resultPromise := Catch(promise, func(err error) error {
		if err != expectedError {
			t.Errorf("Expected error to be %v, but got %v", expectedError, err)
		}
		return err // Should return the error to keep the promise chain accurate.
	})

	<-resultPromise.done

	if resultPromise.err != expectedError {
		t.Errorf("Expected error to be %v, but got %v", expectedError, resultPromise.err)
	}
}

// Der dritte Testfall kann unverändert bleiben

func TestPromiseChain(t *testing.T) {
	promise := NewPromise[int](func(resolve func(int), reject func(error)) {
		time.Sleep(100 * time.Millisecond)
		resolve(1)
	})

	firstResultPromise := Then(promise, func(value int) (int, error) {
		return value + 1, nil
	}, func(err error) error {
		return err
	})

	resultPromise := Then(firstResultPromise, func(value int) (int, error) {
		return value * 2, nil
	}, func(err error) error {
		return err
	})

	<-resultPromise.done

	if resultPromise.err != nil {
		t.Errorf("Promise was rejected with error: %v", resultPromise.err)
	}

	if resultPromise.value != 4 {
		t.Errorf("Promise was resolved with incorrect value: %v, expected %v", resultPromise.value, 4)
	}
}

func BenchmarkPromise(b *testing.B) {
	for i := 0; i < b.N; i++ {
		promise := NewPromise[int](func(resolve func(int), reject func(error)) {
			// Simulieren Sie eine Aufgabe, die einige Zeit in Anspruch nimmt
			time.Sleep(10 * time.Millisecond)
			resolve(42)
		})

		<-promise.done // Warten, bis das Promise-Objekt vollständig gelöst ist
		if promise.err != nil {
			b.Fatal("Promise was rejected with error:", promise.err)
		}
	}
}

func FuzzPromise(f *testing.F) {
	testCases := []struct {
		value     int
		shouldErr bool
	}{
		{value: 42, shouldErr: false},
		{value: -1, shouldErr: true},
	}

	for _, tc := range testCases {
		f.Add(tc.value, tc.shouldErr) // Add test cases to the Fuzz function
	}

	f.Fuzz(func(t *testing.T, value int, shouldErr bool) {
		done := make(chan struct{})
		promise := NewPromise[int](func(resolve func(int), reject func(error)) {
			if shouldErr {
				reject(errors.New("error"))
			} else {
				resolve(value)
			}
		})

		Then(promise, func(data int) (int, error) {
			if data != value {
				t.Errorf("Expected result to be %v, got %v", value, data)
			}
			close(done)
			return data, nil
		}, func(err error) error {
			if !shouldErr {
				t.Errorf("Expected no error, but got: %v", err)
			}
			close(done)
			return err // In real case, we'd expect some error handling here.
		})

		<-done // Wait until Then is resolved or rejected

		if shouldErr && promise.err == nil {
			t.Errorf("Expected an error but did not get one")
		}

		if promise.err != nil && !shouldErr {
			t.Errorf("Did not expect an error but got one: %v", promise.err)
		}
	})
}
